@@ -307,7 +307,7 @@ static void sendAButton(InputInfoPtr pInfo, int button, int mask,
  * @param flags    Flags defining axis attributes: AXIS_INVERT and AXIS_BITWISE
  * @return         Relative change in axis value
  */
-static int getScrollDelta(int current, int old, int wrap, int flags)
+TEST_NON_STATIC int getScrollDelta(int current, int old, int wrap, int flags)
 {
 	int delta;
 
@@ -352,7 +352,7 @@ static int getScrollDelta(int current, int old, int wrap, int flags)
  * @param action_dn    Array index of action to send on scroll down
  * @return             Array index of action that should be performed, or -1 if none.
  */
-static int getWheelButton(int delta, int action_up, int action_dn)
+TEST_NON_STATIC int getWheelButton(int delta, int action_up, int action_dn)
 {
 	if (delta > 0)
 		return action_up;
@@ -815,7 +815,7 @@ void wcmSendEvents(InputInfoPtr pInfo, const WacomDeviceState* ds)
  * @retval SUPPRESS_NONE Process event normally.
  * @retval SUPPRESS_NON_MOTION Suppress all data but motion data.
  */
-static enum WacomSuppressMode
+TEST_NON_STATIC enum WacomSuppressMode
 wcmCheckSuppress(WacomCommonPtr common,
 		 const WacomDeviceState* dsOrig,
 		 WacomDeviceState* dsNew)
@@ -1036,7 +1036,7 @@ static WacomToolPtr findTool(const WacomCommonPtr common,
  *
  * @see normalizePressure
  */
-static int
+TEST_NON_STATIC int
 rebasePressure(const WacomDevicePtr priv, const WacomDeviceState *ds)
 {
 	int min_pressure;
@@ -1065,26 +1065,26 @@ rebasePressure(const WacomDevicePtr priv, const WacomDeviceState *ds)
  * @rebaes
  * @see rebasePressure
  */
-static int
-normalizePressure(const WacomDevicePtr priv, const WacomDeviceState *ds)
+TEST_NON_STATIC int
+normalizePressure(const WacomDevicePtr priv, const int raw_pressure)
 {
 	WacomCommonPtr common = priv->common;
 	double pressure;
-	int p = ds->pressure;
+	int p = raw_pressure;
+	int range_left = common->wcmMaxZ;
 
-	if (p < priv->minPressure)
-	{
-		LogMessageVerbSigSafe(X_ERROR, 0,
-				      "%s: Pressure %d lower than expected minimum %d. This is a bug.\n",
-				      priv->pInfo->name, ds->pressure, priv->minPressure);
-		p = priv->minPressure;
+	if (common->wcmPressureRecalibration) {
+		p -= priv->minPressure;
+		range_left -= priv->minPressure;
 	}
-
 	/* normalize pressure to 0..FILTER_PRESSURE_RES */
-	pressure = xf86ScaleAxis(p - priv->minPressure,
-				 FILTER_PRESSURE_RES, 0,
-				 common->wcmMaxZ - priv->minPressure,
-				 0);
+	if (range_left >= 1)
+		pressure = xf86ScaleAxis(p,
+					 FILTER_PRESSURE_RES, 0,
+					 range_left,
+					 0);
+	else
+		pressure = FILTER_PRESSURE_RES;
 
 	return (int)pressure;
 }
@@ -1097,16 +1097,16 @@ normalizePressure(const WacomDevicePtr priv, const WacomDeviceState *ds)
  * Returns the state of all buttons, but buttons other than button 1 are
  * unmodified.
  */
+#define PRESSURE_BUTTON 1
 static int
-setPressureButton(const WacomDevicePtr priv, const WacomDeviceState *ds)
+setPressureButton(const WacomDevicePtr priv, int buttons, const int pressure)
 {
 	WacomCommonPtr common = priv->common;
-	int button = 1;
-	int buttons = ds->buttons;
+	int button = PRESSURE_BUTTON;
 
 	/* button 1 Threshold test */
 	/* set button1 (left click) on/off */
-	if (ds->pressure < common->wcmThreshold)
+	if (pressure < common->wcmThreshold)
 	{
 		buttons &= ~button;
 		if (priv->oldButtons & button) /* left click was on */
@@ -1114,7 +1114,7 @@ setPressureButton(const WacomDevicePtr priv, const WacomDeviceState *ds)
 			/* don't set it off if it is within the tolerance
 			   and threshold is larger than the tolerance */
 			if ((common->wcmThreshold > THRESHOLD_TOLERANCE) &&
-			    (ds->pressure > common->wcmThreshold - THRESHOLD_TOLERANCE))
+			    (pressure > common->wcmThreshold - THRESHOLD_TOLERANCE))
 				buttons |= button;
 		}
 	}
@@ -1124,6 +1124,45 @@ setPressureButton(const WacomDevicePtr priv, const WacomDeviceState *ds)
 	return buttons;
 }
 
+/*
+ * Broken pen with a broken tip might give high pressure values
+ * all the time. We want to warn about this. To avoid getting
+ * spurious warnings when the tablet is hit quickly will wait
+ * until the device goes out of proximity and check if the minimum
+ * pressure is still above a threshold of 20 percent of the maximum
+ * pressure. Also we make sure the device has seen a sufficient number
+ * of events while in proximity that it had a chance to see decreasing
+ * pressure values.
+ */
+#define LIMIT_LOW_PRESSURE 20 /* percentage of max value */
+#define MIN_EVENT_COUNT 15
+
+static void detectPressureIssue(WacomDevicePtr priv,
+				WacomCommonPtr common,
+				WacomDeviceStatePtr ds)
+{
+	/* pen is just going out of proximity */
+	if (priv->oldProximity && !ds->proximity) {
+
+		int pressureThreshold = common->wcmMaxZ * LIMIT_LOW_PRESSURE / 100;
+		/* check if minPressure has persisted all the time
+		   and is too close to the maximum pressure */
+		if (priv->oldMinPressure > pressureThreshold &&
+		    priv->eventCnt > MIN_EVENT_COUNT)
+			LogMessageVerbSigSafe(
+				X_WARNING, 0,
+				"On %s(%d) a base pressure of %d persists while the pen is in proximity.\n"
+				"\tThis is > %d percent of the maximum value (%d).\n"
+				"\tThis indicates a worn out pen, it is time to change your tool. Also see:\n"
+				"\thttp://sourceforge.net/apps/mediawiki/linuxwacom/index.php?title=Pen_Wear.\n",
+				priv->pInfo->name, priv->serial, priv->minPressure, LIMIT_LOW_PRESSURE, common->wcmMaxZ);
+	} else if (!priv->oldProximity)
+		priv->eventCnt = 0;
+
+	priv->oldMinPressure = priv->minPressure;
+	priv->eventCnt++;
+}
+
 static void commonDispatchDevice(WacomCommonPtr common, unsigned int channel,
 				 const WacomChannelPtr pChannel,
 				 enum WacomSuppressMode suppress)
@@ -1133,6 +1172,7 @@ static void commonDispatchDevice(WacomCommonPtr common, unsigned int channel,
 	WacomDeviceState* ds = &pChannel->valid.states[0];
 	WacomDevicePtr priv = NULL;
 	WacomDeviceState filtered;
+	int raw_pressure = 0;
 
 	/* device_type should have been retrieved and set in the respective
 	 * models, wcmISDV4.c or wcmUSB.c. Once it comes here, something
@@ -1206,10 +1246,51 @@ static void commonDispatchDevice(WacomCommonPtr common, unsigned int channel,
 
 	if ((IsPen(priv) || IsTouch(priv)) && common->wcmMaxZ)
 	{
+		int prev_min_pressure = priv->oldProximity ? priv->minPressure : 0;
+
+		detectPressureIssue(priv, common, &filtered);
+
+		raw_pressure = filtered.pressure;
+		if (!priv->oldProximity)
+			priv->maxRawPressure = raw_pressure;
+
 		priv->minPressure = rebasePressure(priv, &filtered);
-		filtered.pressure = normalizePressure(priv, &filtered);
-		if (IsPen(priv))
-			filtered.buttons = setPressureButton(priv, &filtered);
+
+		filtered.pressure = normalizePressure(priv, filtered.pressure);
+		if (IsPen(priv)) {
+			filtered.buttons = setPressureButton(priv,
+							     filtered.buttons,
+							     filtered.pressure);
+
+			/* Here we run some heuristics to avoid losing button events if the
+			 * pen gets pushed onto the tablet so quickly that the first pressure
+			 * event read is non-zero and is thus interpreted as a pressure bias */
+			if (filtered.buttons & PRESSURE_BUTTON) {
+				/* If we triggered 'normally' reset max pressure to
+				 * avoid to trigger again while this device is in proximity */
+				priv->maxRawPressure = 0;
+			} else if (priv->maxRawPressure) {
+				int norm_max_pressure;
+
+				/* If we haven't triggered normally we record the maximal pressure
+				 * and see if this would have triggered with a lowered bias. */
+				if (priv->maxRawPressure < raw_pressure)
+					priv->maxRawPressure = raw_pressure;
+				norm_max_pressure = normalizePressure(priv, priv->maxRawPressure);
+				filtered.buttons = setPressureButton(priv, filtered.buttons,
+								     norm_max_pressure);
+
+				/* If minPressure is not decrementing any more or a button
+				 * press has been generated or minPressure has just become zero
+				 * reset maxRawPressure to avoid that worn devices
+				 * won't report a button release until going out of proximity */
+				if ((filtered.buttons & PRESSURE_BUTTON &&
+				     priv->minPressure == prev_min_pressure) ||
+				    !priv->minPressure)
+					priv->maxRawPressure = 0;
+
+			}
+		}
 		filtered.pressure = applyPressureCurve(priv,&filtered);
 	}
 	else if (IsCursor(priv) && !priv->oldCursorHwProx)
@@ -1323,7 +1404,7 @@ int wcmInitTablet(InputInfoPtr pInfo, const char* id, float version)
 		return !Success;
 	
 	/* Default threshold value if not set */
-	if (common->wcmThreshold <= 0)
+	if (common->wcmThreshold <= 0 && IsPen(priv))
 	{
 		/* Threshold for counting pressure as a button */
 		common->wcmThreshold = DEFAULT_THRESHOLD;
@@ -1333,19 +1414,17 @@ int wcmInitTablet(InputInfoPtr pInfo, const char* id, float version)
 	}
 
 	/* output tablet state as probed */
-	if (TabletHasFeature(common, WCM_PEN))
-		xf86Msg(X_PROBED, "%s: Wacom %s tablet maxX=%d maxY=%d maxZ=%d "
+	if (IsPen(priv))
+		xf86Msg(X_PROBED, "%s: maxX=%d maxY=%d maxZ=%d "
 			"resX=%d resY=%d  tilt=%s\n",
 			pInfo->name,
-			model->name,
 			common->wcmMaxX, common->wcmMaxY, common->wcmMaxZ,
 			common->wcmResolX, common->wcmResolY,
 			HANDLE_TILT(common) ? "enabled" : "disabled");
-	else
-		xf86Msg(X_PROBED, "%s: Wacom %s tablet maxX=%d maxY=%d maxZ=%d "
+	else if (IsTouch(priv))
+		xf86Msg(X_PROBED, "%s: maxX=%d maxY=%d maxZ=%d "
 			"resX=%d resY=%d \n",
 			pInfo->name,
-			model->name,
 			common->wcmMaxTouchX, common->wcmMaxTouchY,
 			common->wcmMaxZ,
 			common->wcmTouchResolX, common->wcmTouchResolY);
@@ -1443,7 +1522,7 @@ WacomCommonPtr wcmNewCommon(void)
 			/* transmit position if increment is superior */
 	common->wcmRawSample = DEFAULT_SAMPLES;
 			/* number of raw data to be used to for filtering */
-
+	common->wcmPressureRecalibration = 1;
 	return common;
 }
 
@@ -1464,6 +1543,7 @@ void wcmFreeCommon(WacomCommonPtr *ptr)
 					common->serials->serial,
 					common->serials->name);
 
+			free(common->serials->name);
 			next = common->serials->next;
 			free(common->serials);
 			common->serials = next;
