@@ -518,6 +518,7 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 	struct input_absinfo absinfo;
 	unsigned long ev[NBITS(EV_MAX)] = {0};
 	unsigned long abs[NBITS(ABS_MAX)] = {0};
+	unsigned long sw[NBITS(SW_MAX)] = {0};
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common =	priv->common;
 	wcmUSBData* private = common->private;
@@ -744,6 +745,26 @@ int usbWcmGetRanges(InputInfoPtr pInfo)
 	if (!ISBITSET(abs, ABS_MISC))
 		common->wcmProtocolLevel = WCM_PROTOCOL_GENERIC;
 
+	if (ioctl(pInfo->fd, EVIOCGBIT(EV_SW, sizeof(sw)), sw) < 0)
+	{
+		xf86Msg(X_ERROR, "%s: usbProbeKeys unable to ioctl "
+			"sw bits.\n", pInfo->name);
+		return 0;
+	}
+	else if (ISBITSET(sw, SW_MUTE_DEVICE))
+	{
+		common->wcmHasHWTouchSwitch = TRUE;
+
+		memset(sw, 0, sizeof(sw));
+
+		ioctl(pInfo->fd, EVIOCGSW(sizeof(sw)), sw);
+
+		if (ISBITSET(sw, SW_MUTE_DEVICE))
+			common->wcmHWTouchSwitchState = 0;
+		else
+			common->wcmHWTouchSwitchState = 1;
+	}
+
 	usbWcmInitPadState(pInfo);
 
 	return Success;
@@ -881,6 +902,8 @@ static int usbChooseChannel(WacomCommonPtr common, int device_type, unsigned int
 				common->wcmChannel[i].work.proximity = 0;
 				/* dispatch event */
 				wcmEvent(common, i, &common->wcmChannel[i].work);
+				DBG(2, common, "free channels: dropping %u\n",
+						common->wcmChannel[i].work.serial_num);
 			}
 		}
 		DBG(1, common, "device with serial number: %u"
@@ -1205,7 +1228,8 @@ static void usbParseAbsEvent(WacomCommonPtr common,
  *
  * @return The new button mask
  */
-static int mod_buttons(int buttons, int btn, int state)
+TEST_NON_STATIC int
+mod_buttons(int buttons, int btn, int state)
 {
 	int mask;
 
@@ -1464,11 +1488,13 @@ static void usbParseBTNEvent(WacomCommonPtr common,
 			}
 			if (nkeys >= usbdata->npadkeys)
 				change = 0;
-			else if (!ds->device_type) /* expresskey pressed at startup */
-				ds->device_type = PAD_ID;
 	}
 
 	channel->dirty |= change;
+
+	/* expresskey pressed at startup or missing type */
+	if (!ds->device_type && channel->dirty)
+		ds->device_type = PAD_ID;
 }
 
 /**
@@ -1526,11 +1552,11 @@ static int toolTypeToDeviceType(WacomCommonPtr common, int type, int code)
  * @param[in] common
  * @return            A tool type (e.g. STYLUS_ID) associated with the in-prox tool
  */
-static int refreshDeviceType(WacomCommonPtr common)
+static int refreshDeviceType(WacomCommonPtr common, int fd)
 {
 	int device_type = 0;
 	unsigned long keys[NBITS(KEY_MAX)] = { 0 };
-	int rc = ioctl(common->fd, EVIOCGKEY(sizeof(keys)), keys);
+	int rc = ioctl(fd, EVIOCGKEY(sizeof(keys)), keys);
 	int i;
 
 	if (rc == -1) {
@@ -1564,7 +1590,8 @@ static int refreshDeviceType(WacomCommonPtr common)
  *         if last_device_type is not a tool. If all else fails, '0'
  *         is returned.
  */
-static int usbInitToolType(WacomCommonPtr common, const struct input_event *event_ptr,
+static int usbInitToolType(WacomCommonPtr common, int fd,
+                           const struct input_event *event_ptr,
                            int nevents, int last_device_type)
 {
 	int i, device_type = 0;
@@ -1578,7 +1605,7 @@ static int usbInitToolType(WacomCommonPtr common, const struct input_event *even
 		device_type = last_device_type;
 
 	if (!device_type)
-		device_type = refreshDeviceType(common);
+		device_type = refreshDeviceType(common, fd);
 
 	return device_type;
 }
@@ -1601,7 +1628,7 @@ static Bool usbIsTabletToolInProx(int device_type, int proximity)
 
 static void usbDispatchEvents(InputInfoPtr pInfo)
 {
-	int i;
+	int i, c;
 	WacomDeviceState *ds;
 	struct input_event* event;
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
@@ -1612,7 +1639,7 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 
 	DBG(6, common, "%d events received\n", private->wcmEventCnt);
 
-	private->wcmDeviceType = usbInitToolType(common,
+	private->wcmDeviceType = usbInitToolType(common, pInfo->fd,
 	                                         private->wcmEvents,
 	                                         private->wcmEventCnt,
 	                                         dslast.device_type);
@@ -1680,6 +1707,22 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 		if (usbFilterEvent(common, event))
 			continue;
 
+		if (common->wcmHasHWTouchSwitch)
+		{
+			if ((event->type == EV_SW) &&
+			    (event->code == SW_MUTE_DEVICE))
+			{
+				/* touch is disabled when SW_MUTE_DEVICE is set */
+				int touch_enabled = (event->value == 0);
+
+				if (touch_enabled != common->wcmHWTouchSwitchState)
+				/* this property is only set for touch device */
+					wcmUpdateHWTouchProperty(
+						common->wcmTouchDevice,
+						touch_enabled);
+			}
+		}
+
 		/* absolute events */
 		if (event->type == EV_ABS)
 		{
@@ -1722,17 +1765,16 @@ static void usbDispatchEvents(InputInfoPtr pInfo)
 	if (!ds->proximity)
 		private->wcmLastToolSerial = 0;
 
-	/* don't send touch event when touch isn't enabled */
-	if (ds->device_type != TOUCH_ID || common->wcmTouch)
-	{
-		int c;
-		for (c = 0; c < MAX_CHANNELS; c++) {
-			DBG(10, common, "Checking if channel %d is dirty...\n", c);
-			if (common->wcmChannel[c].dirty) {
-				DBG(10, common, "Dirty flag set on channel %d; sending event.\n", c);
-				common->wcmChannel[c].dirty = FALSE;
-				wcmEvent(common, c, &common->wcmChannel[c].work);
-			}
+	for (c = 0; c < MAX_CHANNELS; c++) {
+		ds = &common->wcmChannel[c].work;
+
+		/* walk through all channels */
+		if (common->wcmChannel[c].dirty) {
+			DBG(10, common, "Dirty flag set on channel %d; sending event.\n", c);
+			common->wcmChannel[c].dirty = FALSE;
+			/* don't send touch event when touch isn't enabled */
+			if (ds->device_type != TOUCH_ID || common->wcmTouch)
+				wcmEvent(common, c, ds);
 		}
 	}
 }
