@@ -308,6 +308,16 @@ int wcmDeviceTypeKeys(InputInfoPtr pInfo)
 			break;
 	}
 
+#ifdef INPUT_PROP_DIRECT
+	{
+		int rc;
+		unsigned long prop[NBITS(INPUT_PROP_MAX)] = {0};
+
+		rc = ioctl(pInfo->fd, EVIOCGPROP(sizeof(prop)), prop);
+		if (rc >= 0 && ISBITSET(prop, INPUT_PROP_DIRECT))
+			TabletSetFeature(priv->common, WCM_LCD);
+	}
+#endif
 	if (ISBITSET(common->wcmKeys, BTN_TOOL_PEN))
 		TabletSetFeature(priv->common, WCM_PEN);
 
@@ -706,7 +716,7 @@ int wcmParseSerials (InputInfoPtr pInfo)
 					pInfo->name, serial, name);
 				ser->name = strdup(name);
 			}
-			else ser->name = ""; /*no name yet*/
+			else ser->name = strdup(""); /*no name yet*/
 
 			if (common->serials == NULL)
 				common->serials = ser;
@@ -955,6 +965,12 @@ Bool wcmPreInitParseOptions(InputInfoPtr pInfo, Bool is_primary,
 			common->wcmGestureParameters.wcmTapTime);
 	}
 
+	if (IsStylus(priv) || IsEraser(priv)) {
+		common->wcmPressureRecalibration
+			= xf86SetBoolOption(pInfo->options,
+					    "PressureRecalibration", 1);
+	}
+
 	/* Swap stylus buttons 2 and 3 for Tablet PCs */
 	if (TabletHasFeature(common, WCM_TPC) && IsStylus(priv))
 	{
