@@ -55,6 +55,10 @@
 #define XIGetKnownProperty(prop) 0
 #endif
 
+#ifndef XI86_SERVER_FD
+#define XI86_SERVER_FD 0x20
+#endif
+
 static int wcmDevOpen(DeviceIntPtr pWcm);
 static int wcmReady(InputInfoPtr pInfo);
 static void wcmDevReadInput(InputInfoPtr pInfo);
@@ -107,7 +111,8 @@ void wcmVirtualTabletPadding(InputInfoPtr pInfo)
  *    Initialize logical size and resolution for individual tool.
  ****************************************************************************/
 
-static void wcmInitialToolSize(InputInfoPtr pInfo)
+TEST_NON_STATIC void
+wcmInitialToolSize(InputInfoPtr pInfo)
 {
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
@@ -542,6 +547,22 @@ Bool wcmOpen(InputInfoPtr pInfo)
 	return Success;
 }
 
+/*****************************************************************************
+ * wcmClose --
+ ****************************************************************************/
+
+void wcmClose(InputInfoPtr pInfo)
+{
+	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
+
+	DBG(1, priv, "closing device file\n");
+
+	if (pInfo->fd > -1 && !(pInfo->flags & XI86_SERVER_FD)) {
+		xf86CloseSerial(pInfo->fd);
+		pInfo->fd = -1;
+	}
+}
+
 /*****************************************************************************
  * wcmDevOpen --
  *    Open the physical device and init information structs.
@@ -557,18 +578,17 @@ static int wcmDevOpen(DeviceIntPtr pWcm)
 
 	DBG(10, priv, "\n");
 
+	/* If fd management is done by the server, skip common fd handling */
+	if (pInfo->flags & XI86_SERVER_FD)
+		goto got_fd;
+
 	/* open file, if not already open */
 	if (common->fd_refs == 0)
 	{
 		if ((wcmOpen (pInfo) != Success) || !common->device_path)
 		{
 			DBG(1, priv, "Failed to open device (fd=%d)\n", pInfo->fd);
-			if (pInfo->fd >= 0)
-			{
-				DBG(1, priv, "Closing device\n");
-				xf86CloseSerial(pInfo->fd);
-			}
-			pInfo->fd = -1;
+			wcmClose(pInfo);
 			return FALSE;
 		}
 
@@ -594,6 +614,7 @@ static int wcmDevOpen(DeviceIntPtr pWcm)
 		common->fd_refs++;
 	}
 
+got_fd:
 	/* start the tablet data */
 	if (model->Start && (model->Start(pInfo) != Success))
 		return !Success;
@@ -738,16 +759,17 @@ static void wcmDevClose(InputInfoPtr pInfo)
 	WacomDevicePtr priv = (WacomDevicePtr)pInfo->private;
 	WacomCommonPtr common = priv->common;
 
+	/* If fd management is done by the server, skip common fd handling */
+	if (pInfo->flags & XI86_SERVER_FD)
+		return;
+
 	DBG(4, priv, "Wacom number of open devices = %d\n", common->fd_refs);
 
 	if (pInfo->fd >= 0)
 	{
 		pInfo->fd = -1;
 		if (!--common->fd_refs)
-		{
-			DBG(1, common, "Closing device; uninitializing.\n");
-			xf86CloseSerial (common->fd);
-		}
+			wcmClose(pInfo);
 	}
 }
 
