@@ -43,11 +43,8 @@
 #define LogMessageVerbSigSafe xf86MsgVerb
 #endif
 
-/*****************************************************************************
- * Unit test hack
- ****************************************************************************/
-#ifdef DISABLE_STATIC
-#define static
+#ifndef SW_MUTE_DEVICE
+#define SW_MUTE_DEVICE	0x0e
 #endif
 
 /******************************************************************************
@@ -112,6 +109,9 @@ struct _WacomModule
 /* Open the device with the right serial parmeters */
 extern Bool wcmOpen(InputInfoPtr pInfo);
 
+/* Close the device */
+extern void wcmClose(InputInfoPtr pInfo);
+
 /* device autoprobing */
 char *wcmEventAutoDevProbe (InputInfoPtr pInfo);
 
@@ -175,6 +175,7 @@ extern int wcmDeleteProperty(DeviceIntPtr dev, Atom property);
 extern void InitWcmDeviceProperties(InputInfoPtr pInfo);
 extern void wcmUpdateRotationProperty(WacomDevicePtr priv);
 extern void wcmUpdateSerial(InputInfoPtr pInfo, unsigned int serial, int id);
+extern void wcmUpdateHWTouchProperty(WacomDevicePtr priv, int touch);
 
 /* Utility functions */
 extern Bool is_absolute(InputInfoPtr pInfo);
@@ -191,6 +192,35 @@ enum WacomSuppressMode {
 };
 
 /****************************************************************************/
+
+#ifndef UNIT_TESTS
+
+# define TEST_NON_STATIC static
+
+#else
+
+# define TEST_NON_STATIC
+
+/* For test suite */
+/* xf86Wacom.c */
+extern void wcmInitialToolSize(InputInfoPtr pInfo);
+
+/* wcmConfig.c */
+extern int wcmSetType(InputInfoPtr pInfo, const char *type);
+
+/* wcmCommon.c */
+extern int getScrollDelta(int current, int old, int wrap, int flags);
+extern int getWheelButton(int delta, int action_up, int action_dn);
+extern int rebasePressure(const WacomDevicePtr priv, const WacomDeviceState *ds);
+extern int normalizePressure(const WacomDevicePtr priv, const int raw_pressure);
+extern enum WacomSuppressMode wcmCheckSuppress(WacomCommonPtr common,
+						const WacomDeviceState* dsOrig,
+						WacomDeviceState* dsNew);
+
+/* wcmUSB.c */
+extern int mod_buttons(int buttons, int btn, int state);
+#endif /* UNIT_TESTS */
+
 #endif /* __XF86WACOM_H */
 
 /* vim: set noexpandtab tabstop=8 shiftwidth=8: */
