@@ -303,13 +303,16 @@ struct _WacomDeviceRec
 	int pPressCurve[FILTER_PRESSURE_RES + 1]; /* pressure curve */
 	int nPressCtrl[4];      /* control points for curve */
 	int minPressure;	/* the minimum pressure a pen may hold */
-
+	int oldMinPressure;     /* to record the last minPressure before going out of proximity */
+	unsigned int eventCnt;  /* count number of events while in proximity */
+	int maxRawPressure;     /* maximum 'raw' pressure seen until first button event */
 	WacomToolPtr tool;         /* The common tool-structure for this device */
 
 	int isParent;		/* set to 1 if the device is not auto-hotplugged */
 
 	OsTimerPtr serial_timer; /* timer used for serial number property update */
 	OsTimerPtr tap_timer;   /* timer used for tap timing */
+	OsTimerPtr touch_timer; /* timer used for touch switch property update */
 };
 
 /******************************************************************************
@@ -431,6 +434,8 @@ struct _WacomCommonRec
 	WacomDevicePtr wcmTouchDevice; /* The pointer for pen to access the
 					  touch tool of the same device id */
 	Bool wcmPenInProx;      /* Keep pen in-prox state for touch tool */
+	Bool wcmHasHWTouchSwitch;    /* Tablet has a touch on/off switch */
+	int wcmHWTouchSwitchState;   /* touch event disable/enabled by hardware switch */
 
 	/* These values are in tablet coordinates */
 	int wcmMaxX;                 /* tablet max X value */
@@ -486,6 +491,8 @@ struct _WacomCommonRec
 	int wcmCursorProxoutDistDefault; /* Default max mouse distance for proxy-out */
 	int wcmSuppress;        	 /* transmit position on delta > supress */
 	int wcmRawSample;	     /* Number of raw data used to filter an event */
+	int wcmPressureRecalibration; /* Determine if pressure recalibration of
+					 worn pens should be performed */
 
 	int bufpos;                        /* position with buffer */
 	unsigned char buffer[BUFFER_SIZE]; /* data read from device */
