@@ -203,6 +203,15 @@ static param_t parameters[] =
 		.arg_count = 1,
 		.prop_flags = PROP_FLAG_BOOLEAN
 	},
+	{
+		.name = "HWTouchSwitchState",
+		.desc = "Touch events turned on/off by hardware switch. ",
+		.prop_name = WACOM_PROP_HARDWARE_TOUCH,
+		.prop_format = 8,
+		.prop_offset = 0,
+		.arg_count = 1,
+		.prop_flags = PROP_FLAG_READONLY | PROP_FLAG_BOOLEAN
+	},
 	{
 		.name = "Gesture",
 		.desc = "Turns on/off multi-touch gesture events "
@@ -432,6 +441,15 @@ static param_t parameters[] =
 		.arg_count = 1,
 		.prop_flags = PROP_FLAG_READONLY
 	},
+	{
+		.name = "PressureRecalibration",
+		.desc = "Turns on/off Tablet pressure recalibration",
+		.prop_name = WACOM_PROP_PRESSURE_RECAL,
+		.prop_format = 8,
+		.prop_offset = 0,
+		.arg_count = 1,
+		.prop_flags = PROP_FLAG_BOOLEAN
+	},
 	{
 		.name = "MapToOutput",
 		.desc = "Map the device to the given output. ",
@@ -2791,7 +2809,7 @@ static void test_parameter_number(void)
 	 * deprecated them.
 	 * Numbers include trailing NULL entry.
 	 */
-	assert(ARRAY_SIZE(parameters) == 37);
+	assert(ARRAY_SIZE(parameters) == 39);
 	assert(ARRAY_SIZE(deprecated_parameters) == 17);
 }
 
